from numpy.lib.function_base import quantile
import logging
import numpy as np

import gym.spaces as spaces
from rlberry.agents import AgentWithSimplePolicy
from rlberry.exploration_tools.discrete_counter import DiscreteCounter
from rlberry.agents.dynprog.utils import (
    backward_induction_in_place,
    backward_induction_sd,
)

from algorithms.bayesucbvi.utils import quantile_backward_induction_in_place, quantile_backward_induction_sd 


logger = logging.getLogger(__name__)


class BayesUCBVIAgent(AgentWithSimplePolicy):
    """
    Bayes UCBVI.
    Notes
    -----
    The rewards are supposed to be known


    The recommended policy after all the episodes is computed without
    exploration bonuses.
    Parameters
    ----------
    env : gym.Env
        Environment with discrete states and actions.
    gamma : double, default: 1.0
        Discount factor in [0, 1]. If gamma is 1.0, the problem is set to
        be finite-horizon.
    horizon : int
        Horizon of the objective function. If None and gamma<1, set to
        1/(1-gamma).
    quantile : double, default: 0.85
        Quantile for optimism, controls
        the level of exploration.
    nb_bootstraps: int
        Number of boostrap samples to approximate the quantile.
    reward_free : bool, default: False
        If true, ignores rewards.
    stage_dependent : bool, default: False
        If true, assume that transitions and rewards can change with the stage h.
    References
    ----------
    NA
    """
    name = "BayesUCBVI"

    def __init__(self,
                 env,
                 gamma=1.0,
                 horizon=100,
                 quantile=0.85,
                 nb_bootstraps=100,
                 reward_free=False,
                 stage_dependent=False,
                 **kwargs):
        # init base class
        AgentWithSimplePolicy.__init__(self, env, **kwargs)

        self.gamma = gamma
        self.horizon = horizon
        self.quantile = quantile
        self.nb_bootstraps = nb_bootstraps
        self.reward_free = reward_free
        self.stage_dependent = stage_dependent

        # check environment
        assert isinstance(self.env.observation_space, spaces.Discrete)
        assert isinstance(self.env.action_space, spaces.Discrete)

        # other checks
        assert gamma >= 0 and gamma <= 1.0
        if self.horizon is None:
            assert gamma < 1.0, \
                "If no horizon is given, gamma must be smaller than 1."
            self.horizon = int(np.ceil(1.0 / (1.0 - gamma)))

        # maximum value
        r_range = self.env.reward_range[1] - self.env.reward_range[0]
        if r_range == np.inf or r_range == 0.0:
            logger.warning("{}: Reward range is  zero or infinity. ".format(self.name)
                           + "Setting it to 1.")
            r_range = 1.0

        self.v_max = np.zeros(self.horizon)
        self.v_max[-1] = r_range
        for hh in reversed(range(self.horizon - 1)):
            self.v_max[hh] = r_range + self.gamma * self.v_max[hh + 1]

        # initialize
        self.reset()

    def reset(self, **kwargs):
        H = self.horizon
        S = self.env.observation_space.n
        A = self.env.action_space.n
        B = self.nb_bootstraps

        if self.stage_dependent:
            shape_hsa = (H, S, A)
            shape_hsas = (H, S, A, S+1)
        else:
            shape_hsa = (S, A)
            shape_hsas = (S, A, S+1)

        # (s, a) visit counter
        self.N_sas = np.zeros(shape_hsas)
        self.N_sas[...,-1] += 1
        # MDP estimator
        self.R_hat = np.zeros(shape_hsa)
        self.P_hat = np.zeros(shape_hsas)

        # Value functions
        self.V = np.zeros((H, S+1))
        self.Q = np.zeros((H, S, A))
        # for rec. policy
        self.V_policy = np.zeros((H, S))
        self.Q_policy = np.zeros((H, S, A))

        # Init V
        for hh in range(self.horizon):
            self.V[hh,:] = self.v_max[hh]

        # ep counter
        self.episode = 0

        # useful object to compute total number of visited states & entropy of visited states
        self.counter = DiscreteCounter(self.env.observation_space,
                                       self.env.action_space)

    def policy(self, observation):
        state = observation
        assert self.Q_policy is not None
        return self.Q_policy[0, state, :].argmax()

    def _get_action(self, state, hh=0):
        """ Sampling policy. """
        assert self.Q is not None
        return self.Q[hh, state, :].argmax()
            
    def _update(self, state, action, next_state, reward, hh):
        if self.stage_dependent:
            self.N_sas[hh, state, action, next_state] += 1

            nn = self.N_sas[hh, state, action, :].sum()
            prev_r = self.R_hat[hh, state, action]
            prev_p = self.P_hat[hh, state, action, :]

            self.R_hat[hh, state, action] = (1.0 - 1.0 / nn) * prev_r + reward * 1.0 / nn

            self.P_hat[hh, state, action, :] = (1.0 - 1.0 / nn) * prev_p
            self.P_hat[hh, state, action, next_state] += 1.0 / nn

        else:
            self.N_sas[state, action, next_state] += 1

            nn = self.N_sas[state, action,:].sum()
            prev_r = self.R_hat[state, action]
            prev_p = self.P_hat[state, action, :]

            self.R_hat[state, action] = (1.0 - 1.0 / nn) * prev_r + reward * 1.0 / nn

            self.P_hat[state, action, :] = (1.0 - 1.0 / nn) * prev_p
            self.P_hat[state, action, next_state] += 1.0 / nn

    def _run_episode(self):

        # interact for H steps
        episode_rewards = 0
        state = self.env.reset()
        for hh in range(self.horizon):
            action = self._get_action(state, hh)
            next_state, reward, done, _ = self.env.step(action)
            episode_rewards += reward  # used for logging only

            self.counter.update(state, action)

            if self.reward_free:
                reward = 0.0  # set to zero before update if reward_free

            self._update(state, action, next_state, reward, hh)

            state = next_state
            if done:
                break

        # run backward induction
        H = self.horizon
        B = self.nb_bootstraps
        N_sasb = np.broadcast_to(self.N_sas[...,np.newaxis],self.N_sas.shape+(B,))
        P_weighted = self.rng.gamma(N_sasb)
        P_weighted = P_weighted/P_weighted.sum(-2,keepdims = True)
        if self.stage_dependent:
            quantile_backward_induction_sd(
                self.Q,
                self.V,
                self.R_hat,#np.broadcast_to(self.env.R[np.newaxis,...],(H,)+self.env.R.shape),
                P_weighted,
                self.gamma,
                self.v_max[0],
                self.quantile)
        else:
            quantile_backward_induction_in_place(
                self.Q,
                self.V,
                self.R_hat,#self.env.R,
                P_weighted,
                self.horizon,
                self.gamma,
                self.v_max[0],
                self.quantile)

        # update info
        self.episode += 1

        # writer
        if self.writer is not None:
            self.writer.add_scalar("episode_rewards", episode_rewards, self.episode)
            self.writer.add_scalar("n_visited_states", self.counter.get_n_visited_states(), self.episode)

        # return sum of rewards collected in the episode
        return episode_rewards

    def fit(self, budget: int, **kwargs):
        del kwargs
        n_episodes_to_run = budget
        count = 0
        while count < n_episodes_to_run:
            self._run_episode()
            count += 1

        # compute Q function for the recommended policy
        if self.stage_dependent:
            backward_induction_sd(
                self.Q_policy,
                self.V_policy,
                self.R_hat,
                self.P_hat,
                self.gamma,
                self.v_max[0])
        else:
            backward_induction_in_place(
                self.Q_policy,
                self.V_policy,
                self.R_hat,
                self.P_hat,
                self.horizon,
                self.gamma,
                self.v_max[0])