from rlberry.utils.jit_setup import numba_jit
import numpy as np


@numba_jit
def quantile_backward_induction_in_place(Q, V, R, P, horizon, gamma=1.0, vmax=np.inf, quantile = 0.95):
    """
    Backward induction to compute Q and V functions in
    the finite horizon setting.
    Takes as input the arrays where to store Q and V.
    Parameters
    ----------
    Q:  numpy.ndarray
        array of shape (horizon, S, A) where to store the Q function
    V:  numpy.ndarray
        array of shape (horizon, S+1) where to store the V function
    R : numpy.ndarray
        array of shape (S, A, B) contaning the rewards, where S is the number
        of states and A is the number of actions Bnumber of bootstraps
    P : numpy.ndarray
        array of shape (S, A, S+1,B) such that P[s,a,ns,b] is the probability of
        arriving at ns by taking action a in state s for the bootstrap b
    horizon : int
        problem horizon
    gamma : double
        discount factor, default = 1.0
    vmax : double
        maximum possible value in V
        default = np.inf
    """
    S, A, B = R.shape
    _q_aa = np.zeros(B)
    for hh in range(horizon - 1, -1, -1):
        for ss in range(S):
            max_q = -np.inf
            for aa in range(A):
                for bb in range(B):
                  _q_aa[bb] = R[ss, aa, bb]
                  if hh < horizon - 1:
                      for ns in range(S+1):
                        _q_aa[bb] += gamma * P[ss, aa, ns, bb] * V[hh + 1, ns]
                q_aa = np.quantile(_q_aa, quantile)
                if q_aa > max_q:
                    max_q = q_aa
                Q[hh, ss, aa] = q_aa
            V[hh, ss] = max_q
            if V[hh, ss] > vmax:
                V[hh, ss] = vmax


@numba_jit
def quantile_backward_induction_sd(Q, V, R, P, gamma=1.0, vmax=np.inf, quantile = 0.95):
    """
    In-place implementation of backward induction to compute Q and V functions
    in the finite horizon setting.
    Assumes R and P are stage-dependent.
    Parameters
    ----------
    Q:  numpy.ndarray
        array of shape (H, S, A) where to store the Q function
    V:  numpy.ndarray
        array of shape (H, S+1) where to store the V function
    R : numpy.ndarray
        array of shape (H, S, A,B) contaning the rewards, where S is the number
        of states and A is the number of actions
    P : numpy.ndarray
        array of shape (H, S, A, S+1,B) such that P[h, s, a, ns,b] is the probability of
        arriving at ns by taking action a in state s at stage h in bootstrap b.
    gamma : double, default: 1.0
        discount factor
    vmax : double, default: np.inf
        maximum possible value in V
    quantile: the quantile for optimism
    """
    H, S, A, B= R.shape
    _q_aa = np.zeros(B)
    for hh in range(H - 1, -1, -1):
        for ss in range(S):
            max_q = -np.inf
            for aa in range(A):
                for bb in range(B):
                    _q_aa[bb] = R[hh, ss, aa, bb]
                    if hh < H - 1:
                        for ns in range(S+1):
                            _q_aa[bb] += gamma * P[hh, ss, aa, ns,bb] * V[hh + 1, ns]
                q_aa = np.quantile(_q_aa, quantile)
                if q_aa > max_q:
                    max_q = q_aa
                Q[hh, ss, aa] = q_aa
            V[hh, ss] = max_q
            # clip V
            if V[hh, ss] > vmax:
                V[hh, ss] = vmax


@numba_jit
def update_value_and_get_action(state,
                                hh,
                                V,
                                R_hat,
                                P_hat,
                                gamma,
                                v_max,
                                quantile):
  """
  state : int
  hh : int
  V : np.ndarray
      shape (H, S+1)
  R_hat : np.ndarray
      shape (S, A, B)
  P_hat : np.ndarray
      shape (S, A, S+1, B)
  gamma : double
  v_max : np.ndarray
      shape (H,)
  """
  H = V.shape[0]
  S, A, B = R_hat.shape
  best_action = 0
  max_val = 0
  previous_value = V[hh, state]
  _q_aa = np.zeros(B)

  for aa in range(A):
    for bb in range(B):
      _q_aa[bb] = R_hat[state, aa,bb]
      if hh < H - 1:
        for sn in range(S+1):
          _q_aa[bb] += gamma * P_hat[state, aa, sn, bb] * V[hh + 1, sn]
    q_aa = np.quantile(_q_aa, quantile)
    if aa == 0 or q_aa > max_val:
      max_val = q_aa
      best_action = aa

  V[hh, state] = max_val
  V[hh, state] = min(v_max[hh], V[hh, state])
  V[hh, state] = min(previous_value, V[hh, state])

  return best_action


@numba_jit
def update_value_and_get_action_sd(state,
                                   hh,
                                   V,
                                   R_hat,
                                   P_hat,
                                   B_sa,
                                   gamma,
                                   v_max,
                                   quantile):
  """
  state : int
  hh : int
  V : np.ndarray
      shape (H, S+1)
  R_hat : np.ndarray
      shape (H, S, A, B)
  P_hat : np.ndarray
      shape (H, S, A, S+1,B)
  gamma : double
  v_max : np.ndarray
      shape (H,)
  """
  H = V.shape[0]
  S, A, B = R_hat.shape[-3:]
  best_action = 0
  max_val = 0
  previous_value = V[hh, state]
  _q_aa = np.zeros(B)

  for aa in range(A):
    for bb in range(B):
      _q_aa[bb] = R_hat[hh, state, aa, bb]
      if hh < H - 1:
          for sn in range(S):
              _q_aa[bb] += gamma * P_hat[hh, state, aa, sn, bb] * V[hh + 1, sn]
      q_aa = np.quantile(_q_aa, quantile)
      if aa == 0 or q_aa > max_val:
          max_val = q_aa
          best_action = aa

  V[hh, state] = max_val
  V[hh, state] = min(v_max[hh], V[hh, state])
  V[hh, state] = min(previous_value, V[hh, state])

  return best_action